{
  "type": "record",
  "name": "BrowsePaths",
  "namespace": "com.linkedin.pegasus2avro.common",
  "doc": "Shared aspect containing Browse Paths to be indexed for an entity.",
  "fields": [
    {
      "name": "paths",
      "type": {
        "type": "array",
        "items": "string"
      },
      "doc": "A list of valid browse paths for the entity.\n\nBrowse paths are expected to be forward slash-separated strings. For example: 'prod/snowflake/datasetName'",
      "Searchable": {
        "/*": {
          "fieldName": "browsePaths",
          "fieldType": "BROWSE_PATH"
        }
      }
    }
  ],
  "Aspect": {
    "name": "browsePaths"
  }
}